typedef struct Command {
    struct Command *next;

    union {
        struct {
            Char *key;
        } get;

        struct {
            Char *key;
            Char *value;
        } put;

        struct {
            Char *key;
        } del;
    };

    enum Command_Kind {
        Command_Kind_NONE,
        Command_Kind_GET,
        Command_Kind_PUT,
        Command_Kind_DEL,
    } kind;

    Sint fd;
} Command;

global_variable Command *global_command_first, *global_command_last, *global_command_divider;
global_variable pthread_mutex_t global_command_lock = PTHREAD_MUTEX_INITIALIZER;
global_variable pthread_cond_t global_command_cond_var = PTHREAD_COND_INITIALIZER;

internal_function
void commandInitBuffer (void)
{
    global_command_first = calloc(1, sizeof(*global_command_first));
    global_command_last = global_command_first;
    global_command_divider = global_command_first;

    return;
}

internal_function
void commandEnqueue (Command c)
{
    global_command_last->next = calloc(1, sizeof(*(global_command_last->next)));
    *(global_command_last->next) = c;
    global_command_last = global_command_last->next;

    while (global_command_first != global_command_divider) {
        Command *temp = global_command_first;
        global_command_first = global_command_first->next;
        free(temp);
    }

    pthread_cond_broadcast(&global_command_cond_var);

    return;
}

internal_function
void commandDequeue (Command *c)
{
    pthread_mutex_lock(&global_command_lock);

    while (global_command_divider == global_command_last) {
        pthread_cond_wait(&global_command_cond_var, &global_command_lock);
    }

    *c = *(global_command_divider->next);
    global_command_divider = global_command_divider->next;

    pthread_mutex_unlock(&global_command_lock);
}

internal_function
void* commandProcessLoop (void *arg)
{
    Thread_Pool_Metadata *data = arg;
    XML_Dump_Metadata *xml_dump = data->xml_dump;

    while (true) {
        Command command = {0};
        commandDequeue(&command);
        switch (command.kind) {
            case Command_Kind_GET: {
                Char *error = NULL;
                U64 hash = hashFNV1a(command.get.key);

//                pthread_mutex_lock(&data_lock);
                Char *value = storageGetFileData(hash, command.get.key, &error);
//                pthread_mutex_unlock(&data_lock);

                Result result = {0};

                if (value == NULL) {
                    result = (Result){.kind = Result_Kind_ERROR,
                                      .fd = command.fd,
                                      .error.error = error};
                } else {
                    result = (Result){.kind = Result_Kind_GET,
                                      .fd = command.fd,
                                      .get.key = command.get.key,
                                      .get.value = value};
                }

                resultEnqueue(result);
            } break;

            case Command_Kind_PUT: {
                Char *error = NULL;
                U64 hash = hashFNV1a(command.put.key);

//                pthread_mutex_lock(&data_lock);
                B32 success = storageSetFileData(hash, command.put.key, command.put.value,
                                                 &error);
//                pthread_mutex_unlock(&data_lock);

                Result result = {0};

                if (success == false) {
                    result = (Result){.kind = Result_Kind_ERROR,
                                      .fd = command.fd,
                                      .error.error = error};
                } else {
                    pthread_cancel(xml_dump->thread);
                    pthread_join(xml_dump->thread, NULL);
                    fclose(xml_dump->file);
                    xml_dump->file = fopen(xml_dump->file_path, "w");
                    pthread_create(&(xml_dump->thread), NULL, &dump, xml_dump);

                    result = (Result){.kind = Result_Kind_PUT,
                                      .fd = command.fd};
                }

                resultEnqueue(result);
            } break;

            case Command_Kind_DEL: {
                Char *error = NULL;
                U64 hash = hashFNV1a(command.del.key);

//                pthread_mutex_lock(&data_lock);
                B32 success = storageDeleteFileData(hash, command.del.key, &error);
//                pthread_mutex_unlock(&data_lock);

                Result result = {0};

                if (success == false) {
                    result = (Result){.kind = Result_Kind_ERROR,
                                      .fd = command.fd,
                                      .error.error = error};
                } else {
                    pthread_cancel(xml_dump->thread);
                    pthread_join(xml_dump->thread, NULL);
                    fclose(xml_dump->file);
                    xml_dump->file = fopen(xml_dump->file_path, "w");
                    pthread_create(&(xml_dump->thread), NULL, &dump, xml_dump);

                    result = (Result){.kind = Result_Kind_DEL,
                                      .fd = command.fd};
                }

                resultEnqueue(result);
            } break;
            default: {
            } break;
        }
    }

    return NULL;
}
