/* The network code was inspired by the following blog post:
 *    https://eli.thegreenplace.net/2017/concurrent-servers-part-3-event-driven
 */

#include "aux.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/epoll.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <pthread.h>
#include <dirent.h>

#if 0
# define log(...) do {printf("Log: " __VA_ARGS__); fflush(stdout);} while(0)
#else
# define log(...)
#endif

typedef struct XML_Dump_Metadata {
    FILE *file;
    Char *file_path;
    pthread_t thread;
} XML_Dump_Metadata;

typedef struct Thread_Pool_Metadata {
    XML_Dump_Metadata *xml_dump;
} Thread_Pool_Metadata;

#include "xml.h"

#include "dump.c"
#include "cache.c"
#include "storage.c"
#include "result_buffer.c"
#include "command_buffer.c"

#define MAX_SOCKET_CONNECTIONS_REQUEST 64
#define MAX_EPOLL_EVENTS 1024

typedef struct Input_Resume {
    Char *buffer;
    Size buffer_len;
    int fd;
} Input_Resume;

typedef struct Output_Resume {
    Char *output;
    Size output_pos;
    Size output_len;
    Sint fd;
} Output_Resume;

Sint main (Sint argc, Char *argv[])
{
    Char *port = NULL;
    Size thread_pool_size = 0;
    Size cache_set_count = 0;
    Size cache_associativity = 0;

    for (Size i = 1; i < (Size)argc; i++) {
        Size arg_pos = 0;
        if ((arg_pos = strprefix("-port=", argv[i]))) {
            port = &(argv[i][arg_pos]);
        } else if ((arg_pos = strprefix("-threadPoolSize=", argv[i]))) {
            thread_pool_size = (Size)strtol(&(argv[i][arg_pos]), NULL, 0);
        } else if ((arg_pos = strprefix("-numSetsInCache=", argv[i]))) {
            cache_set_count = nextPowerOf2((Size)strtol(&(argv[i][arg_pos]), NULL, 0));
        } else if ((arg_pos = strprefix("-sizeOfSet=", argv[i]))) {
            cache_associativity = (Size)strtol(&(argv[i][arg_pos]), NULL, 0);
        }
    }

    if (port == NULL) port = "8080";
    if (thread_pool_size == 0) thread_pool_size = 100;
    if (cache_set_count == 0) cache_set_count = 32;
    if (cache_associativity == 0) cache_associativity = 2;

    cacheInit(cache_set_count, cache_associativity);

    struct stat st = {0};
    if(stat("data_store", &st) == -1) {
        if(mkdir("data_store", 0777) < 0) {
            fprintf(stderr, "Error: Cannot create directory %s because: ",
                   "data_store");
            perror("");
            exit(-1);
        }
    }

    if(chdir("data_store") < 0) {
        fprintf(stderr, "Error: Cannot change to directory %s because: ",
               "data_store");
        perror("");
        exit(-1);
    }

    resultInitBuffer();
    commandInitBuffer();

    XML_Dump_Metadata xml_dump = {.file_path = "kvstore.xml"};
    xml_dump.file = fopen(xml_dump.file_path, "w");
    pthread_create(&(xml_dump.thread), NULL, &dump, &xml_dump);

    pthread_t *threads = calloc(thread_pool_size, sizeof(*threads));
    Thread_Pool_Metadata thread_metadata = {.xml_dump = &xml_dump};
    for (Size i = 0; i < thread_pool_size; i++) {
        pthread_create(&threads[i], NULL, &commandProcessLoop, &thread_metadata);
    }

    Hashmap output_map = hmCreate(memCRT, 1024);
    Hashmap input_map = hmCreate(memCRT, 1024);
    Hashmap busy_map = hmCreate(memCRT, 1024);

    // NOTE(naman): Create a socket for IPv4 and TCP.
    Sint sock_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (sock_fd < 0) {
        perror("ERROR opening socket");
        exit(-1);
    }

    // NOTE(naman): This helps avoid spurious EADDRINUSE when the previous instance of this
    // server died.
    int opt = 1;
    if (setsockopt(sock_fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
        perror("setsockopt");
        exit(-1);
    }

    // NOTE(naman): Get actual internet address to bind to using IPv4 and TCP,
    // and listening passively
    struct addrinfo hints = {.ai_family = AF_INET,
                             .ai_socktype = SOCK_STREAM,
                             .ai_flags = AI_PASSIVE};
    struct addrinfo *addrinfo = NULL;
    Sint s = getaddrinfo(NULL, port, &hints, &addrinfo);
    if (s != 0) {
        fprintf(stderr, "Error: getaddrinfo: %s\n", gai_strerror(s));
        exit(-1);
    }

    // NOTE(naman): Assign an address to the socket
    if (bind(sock_fd, addrinfo->ai_addr, addrinfo->ai_addrlen) != 0) {
        perror("bind()");
        exit(-1);
    }

    // NOTE(naman): Start listening for incoming connections
    if (listen(sock_fd, MAX_SOCKET_CONNECTIONS_REQUEST) != 0) {
        perror("listen()");
        exit(-1);
    }

    // NOTE(naman): Set the socket as non-blocking
    int flags = fcntl(sock_fd, F_GETFL, 0);
    if (flags == -1) {
        perror("fcntl F_GETFL");
        exit(-1);
    }
    if (fcntl(sock_fd, F_SETFL, flags | O_NONBLOCK) == -1) {
        perror("fcntl F_SETFL O_NONBLOCK");
        exit(-1);
    }


    // NOTE(naman): Create an epoll instance, with no flags set.
    int epoll_fd = epoll_create1(0);
    if (epoll_fd < 0) {
        perror("epoll_create1");
        exit(-1);
    }

    // NOTE(naman): Add the socket's file descriptor to the epoll set
    struct epoll_event accept_event = {.data.fd = sock_fd,
                                       .events = EPOLLIN};
    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock_fd, &accept_event) < 0) {
        perror("epoll_ctl EPOLL_CTL_ADD");
        exit(-1);
    }

    struct epoll_event* events = calloc(MAX_EPOLL_EVENTS, sizeof(struct epoll_event));
    if (events == NULL) {
        fprintf(stderr, "Unable to allocate memory for epoll_events");
        exit(-1);
    }

    printf("Log: Waiting for connection on port %s...\n", port);

    while (true) {
        // NOTE(naman): Get the fd's that are ready
        int nready = epoll_wait(epoll_fd, events, MAX_EPOLL_EVENTS, -1);

        for (int i = 0; i < nready; i++) {
            if (events[i].events & EPOLLERR) {
                perror("epoll_wait returned EPOLLERR");
                exit(-1);
            }

            if (events[i].data.fd == sock_fd) {
                // NOTE(naman): A new client is connecting.
                struct sockaddr_in peer_addr = {0};
                socklen_t peer_addr_len = sizeof(peer_addr);
                int accept_fd = accept(sock_fd, (struct sockaddr*)&peer_addr,
                                       &peer_addr_len);

                if (accept_fd < 0) {
                    if (errno == EAGAIN || errno == EWOULDBLOCK) {
                        // This can happen due to the nonblocking socket mode; in this
                        // case don't do anything, but print a notice (since these events
                        // are extremely rare and interesting to observe...)
                        fprintf(stderr, "accept() returned %s\n",
                                errno == EAGAIN ? "EAGAIN" : "EWOULDBLOCK");
                    } else {
                        perror("accept");
                        exit(-1);
                    }
                } else {
                    printf("Log: Connection made: client_fd=%d\n", accept_fd);

                    // NOTE(naman): Set the socket as non-blocking
                    int flags = fcntl(sock_fd, F_GETFL, 0);
                    if (flags == -1) {
                        perror("fcntl F_GETFL");
                        exit(-1);
                    }
                    if (fcntl(sock_fd, F_SETFL, flags | O_NONBLOCK) == -1) {
                        perror("fcntl F_SETFL O_NONBLOCK");
                        exit(-1);
                    }

                    // NOTE(naman): Add the new file descriptor to the epoll set
                    struct epoll_event event = {.data.fd = accept_fd,
                                                .events = EPOLLIN};
                    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, accept_fd, &event) < 0) {
                        perror("epoll_ctl EPOLL_CTL_ADD");
                        exit(-1);
                    }
                }
            } else {
                // A peer socket is ready.
                if (events[i].events & EPOLLIN) {
                    // Ready for reading.
                    int fd = events[i].data.fd;

                    char *buffer = calloc(KiB(300), sizeof(*buffer));
                    int len = read(fd, buffer, KiB(300) - 1);

                    if (len == 0) {
                        close(fd);
                        continue;
                    }

                    B32 is_busy = hmLookupI(&busy_map, fd);
                    Input_Resume *ir = (Input_Resume*)hmLookupI(&input_map, fd);

                    if (is_busy && (ir == NULL)) {
                        continue;
                    }

                    hmInsertI(&busy_map, fd, true);

                    if (ir == NULL) {
                        ir = calloc(1, sizeof(*ir));
                        ir->buffer = calloc(KiB(300), sizeof(*(ir->buffer)));
                        ir->fd = fd;
                        hmInsertI(&input_map, fd, (Uptr)ir);
                    }

                    strcat(ir->buffer, buffer);
                    ir->buffer_len += len;
                    free(buffer);

                    if (strsuffix(ir->buffer, "</KVMessage>\n") && (ir->buffer_len <= KiB(299))) {
                        XML_Message msg = xmlParseMessage(ir->buffer);

                        switch (msg.kind) {
                            case XML_Message_Kind_GET: {
                                Command command = {.fd = fd};

                                command.kind = Command_Kind_GET;
                                command.get.key = msg.key;

                                commandEnqueue(command);
                            } break;

                            case XML_Message_Kind_PUT: {
                                Command command = {.fd = fd};

                                command.kind = Command_Kind_PUT;
                                command.put.key = msg.key;
                                command.put.value = msg.value;

                                commandEnqueue(command);
                            } break;

                            case XML_Message_Kind_DELETE: {
                                Command command = {.fd = fd};

                                command.kind = Command_Kind_DEL;
                                command.del.key = msg.key;

                                commandEnqueue(command);
                            } break;

                            case XML_Message_Kind_SYSTEM_ERROR: {
                                Char *output = xmlCreateMessage(XML_Message_Kind_RESP_ERROR,
                                                                NULL, NULL, msg.error);
                                Size output_len = strlen(output);

                                Output_Resume *or = calloc(1, sizeof(*or));
                                or->fd = fd;
                                or->output = output;
                                or->output_pos = 0;
                                or->output_len = output_len;
                                hmInsertI(&output_map, fd, (Uptr)or);
                            } break;

                            default: {
                            } break;
                        }

                        hmRemoveI(&input_map, fd);
                        free(ir->buffer);
                        free(ir);

                        struct epoll_event event = {.data.fd = fd,
                                                    .events = EPOLLIN | EPOLLOUT};
                        if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, fd, &event) < 0) {
                            perror("epoll_ctl EPOLL_CTL_MOD");
                            exit(-1);
                        }
                    }
                } else if (events[i].events & EPOLLOUT) {
                    // Writing into fd in which we previously were not able to finish writing to
                    int fd = events[i].data.fd;

                    Output_Resume *or = hmLookup(&output_map, (void*)(Uptr)fd);
                    if (or == NULL) {
                        // fprintf(stderr, "hmLookup returned NULL\n");
                        continue;
                    } else {
                        Char *output = or->output;
                        Size output_len = or->output_len;
                        Size output_pos = or->output_pos;

                        int nsent = write(or->fd, output + output_pos, output_len - output_pos);

                        if (nsent == -1) {
                            if (errno == EAGAIN || errno == EWOULDBLOCK) {
                                // Try next time
                            } else {
                                perror("write() failed\n");
                                exit(-1);
                            }
                        } else if ((Size)nsent < (output_len - output_pos)) {
                            or->output_pos += nsent;
                        } else {
                            sbufDelete(output);
                            hmRemove(&output_map, (void*)(Uptr)fd);
                            hmRemoveI(&busy_map, fd);
                            free(or);

                            struct epoll_event event = {.data.fd = fd,
                                                        .events = EPOLLIN};
                            if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD,
                                          or->fd, &event) < 0) {
                                perror("epoll_ctl EPOLL_CTL_MOD");
                                exit(-1);
                            }
                        }
                    }
                }
            }
        }

        Result result = {0};
        while (resultDequeue(&result)) {
            Char *key = NULL, *value = NULL, *error = NULL;
            XML_Message_Kind xml_kind = XML_Message_Kind_NONE;

            switch (result.kind) {
                case Result_Kind_GET: {
                    key = result.get.key;
                    value = result.get.value;
                    xml_kind = XML_Message_Kind_RESP_GET;
                } break;
                case Result_Kind_PUT: {
                    xml_kind = XML_Message_Kind_RESP_PUT_DELETE;
                } break;
                case Result_Kind_DEL: {
                    xml_kind = XML_Message_Kind_RESP_PUT_DELETE;
                } break;
                case Result_Kind_ERROR: {
                    error = result.error.error;
                    xml_kind = XML_Message_Kind_RESP_ERROR;
                } break;
                default: {
                } break;
            }

            Char *output = xmlCreateMessage(xml_kind, key, value, error);
            Size output_len = strlen(output);

            int nsent = write(result.fd, output, output_len);

            if (nsent == -1) {
                if (errno == EAGAIN || errno == EWOULDBLOCK) {
                    Output_Resume *or = calloc(1, sizeof(*or));
                    or->fd = result.fd;
                    or->output = output;
                    or->output_pos = 0;
                    or->output_len = output_len;
                    hmInsertI(&output_map, result.fd, (Uptr)or);

                    struct epoll_event event = {.data.fd = result.fd,
                                                .events = EPOLLIN | EPOLLOUT};
                    if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, result.fd, &event) < 0) {
                        perror("epoll_ctl EPOLL_CTL_MOD");
                        exit(-1);
                    }
                } else {
                    perror("write() failed");
                    exit(-1);
                }
            } else if ((Size)nsent < output_len) {
                Output_Resume *or = calloc(1, sizeof(*or));
                or->fd = result.fd;
                or->output = output;
                or->output_pos = nsent;
                or->output_len = output_len;
                hmInsertI(&output_map, result.fd, (Uptr)or);

                struct epoll_event event = {.data.fd = result.fd,
                                            .events = EPOLLIN | EPOLLOUT};
                if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, result.fd, &event) < 0) {
                    perror("epoll_ctl EPOLL_CTL_MOD");
                    exit(-1);
                }
            } else {
                struct epoll_event event = {.data.fd = result.fd,
                                            .events = EPOLLIN};
                if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD,
                              result.fd, &event) < 0) {
                    perror("epoll_ctl EPOLL_CTL_MOD");
                    exit(-1);
                }

                sbufDelete(output);
                hmRemoveI(&busy_map, result.fd);
            }

            switch (result.kind) {
                case Result_Kind_GET: {
                    free(result.get.key);
                    free(result.get.value);
                } break;
                case Result_Kind_PUT: {
                } break;
                case Result_Kind_DEL: {
                } break;
                case Result_Kind_ERROR: {
                    free(result.error.error);
                } break;
                default: {
                } break;
            }
        }
    }

    return 0;
}
